"""Tests for the BaseScheduler class."""

from __future__ import annotations

import textwrap

import pytest

from .helpers import MockScheduler


def test_base_scheduler_job_script() -> None:
    """Test the BaseScheduler.job_script method."""
    s = MockScheduler(
        cores=4,
        extra_scheduler=["--exclusive=user", "--time=1"],
        extra_env_vars=["TMPDIR='/scratch'", "PYTHONPATH='my_dir:$PYTHONPATH'"],
        extra_script="echo 'YOLO'",
        executor_type="mpi4py",
    )
    job_script = s.job_script(options={})
    log_fname = s.log_fname("${NAME}")
    assert job_script == textwrap.dedent(
        f"""\
        #!/bin/bash
        #MOCK --cores 4
        ##MOCK --exclusive=user
        ##MOCK --time=1

        export TMPDIR='/scratch'
        export PYTHONPATH='my_dir:$PYTHONPATH'
        export EXECUTOR_TYPE=mpi4py
        export MKL_NUM_THREADS=1
        export OPENBLAS_NUM_THREADS=1
        export OMP_NUM_THREADS=1
        export NUMEXPR_NUM_THREADS=1

        echo 'YOLO'

        {s.mpiexec_executable} \\
            -n 4 {s.python_executable} \\
            -m mpi4py.futures {s.launcher} \\
            --log-fname {log_fname} \\
            --job-id ${{JOB_ID}} \\
            --name ${{NAME}}
        """,
    )


def test_base_scheduler() -> None:
    """Test that the base scheduler is set correctly."""
    s = MockScheduler(cores=4)
    assert s.cores == 4


def test_queue() -> None:
    """Test the queue method of MockScheduler."""
    scheduler = MockScheduler(cores=1)
    scheduler.start_job("test_job1")
    scheduler.start_job("test_job2")

    queue_info = scheduler.queue()
    assert len(queue_info) == 2
    assert queue_info["0"]["job_name"] == "test_job1"
    assert queue_info["1"]["job_name"] == "test_job2"


def test_cancel() -> None:
    """Test the cancel method of MockScheduler."""
    scheduler = MockScheduler(cores=2)
    scheduler.start_job("test_job1")
    scheduler.start_job("test_job2")
    assert len(scheduler.queue()) == 2
    scheduler.cancel(["test_job1"])

    queue_info = scheduler.queue()
    assert len(queue_info) == 1
    assert queue_info["1"]["job_name"] == "test_job2"


def test_update_queue() -> None:
    """Test the update_queue method of MockScheduler."""
    scheduler = MockScheduler(cores=2)
    scheduler.start_job("test_job1")
    scheduler.start_job("test_job2")
    j1, _ = scheduler.job_names_to_job_ids("test_job1", "test_job2")
    scheduler.update_queue("test_job1", "COMPLETED")

    queue_info = scheduler.queue()
    assert len(queue_info) == 2
    assert queue_info[j1]["status"] == "COMPLETED"


def test_ipyparallel() -> None:
    """Test that the ipyparallel raises."""
    s = MockScheduler(cores=1, executor_type="ipyparallel")
    with pytest.raises(
        ValueError,
        match="`ipyparalllel` uses 1 cores of the `adaptive.Runner`",
    ):
        s._executor_specific("NAME", {})


def test_getstate_setstate() -> None:
    """Test that the getstate and setstate methods work."""
    s = MockScheduler(cores=4)
    state = s.__getstate__()
    s2 = MockScheduler(cores=2)
    s2.__setstate__(state)
    assert s.cores == s2.cores


def test_base_scheduler_ipyparallel() -> None:
    """Test the BaseScheduler.job_script method."""
    s = MockScheduler(
        cores=4,
        extra_scheduler=["--exclusive=user", "--time=1"],
        extra_env_vars=["TMPDIR='/scratch'", "PYTHONPATH='my_dir:$PYTHONPATH'"],
        extra_script="echo 'YOLO'",
        executor_type="ipyparallel",
    )
    ipy = s._executor_specific("TEST", {"--save-dataframe": None, "--log-interval": 10})
    log_fname = s.log_fname("TEST")
    print(ipy)
    assert (
        ipy.strip()
        == textwrap.dedent(
            f"""\
            profile=adaptive_scheduler_${{JOB_ID}}

            echo "Creating profile ${{profile}}"
            ipython profile create ${{profile}}

            echo "Launching controller"
            ipcontroller --ip="*" --profile=${{profile}} --log-to-file &
            sleep 10

            echo "Launching engines"
            mpiexec \\
                -n 3 \\
                ipengine \\
                --profile=${{profile}} \\
                --mpi \\
                --cluster-id='' \\
                --log-to-file &

            echo "Starting the Python script"
            {s.python_executable} {s.launcher} \\
                --profile ${{profile}} \\
                --n 3 \\
                --log-fname {log_fname} \\
                --job-id ${{JOB_ID}} \\
                --name TEST \\
                --save-dataframe \\
                --log-interval 10
            """,
        ).strip()
    )


def test_callable_scheduler_arguments() -> None:
    """Test that the scheduler arguments can be callables."""
    s = MockScheduler(
        cores=(4, lambda: 2),
        executor_type=(lambda: "ipyparallel", "mpi4py"),
        num_threads=(lambda: 2, 1),
        extra_scheduler=(
            ["--exclusive=user", "--time=1"],
            lambda: ["--exclusive=user", "--time=2"],
        ),
        extra_env_vars=(lambda: ["from=func"], ["from=static"]),
        extra_script=(lambda: "echo 'func'", "echo 'static'"),
    )

    js0 = s.job_script(options={}, index=0)
    js1 = s.job_script(options={}, index=1)
    assert js0 != js1
    extra_scheduler0 = s.extra_scheduler(index=0)
    extra_scheduler1 = s.extra_scheduler(index=1)
    assert extra_scheduler0 == "##MOCK --exclusive=user\n##MOCK --time=1"
    assert extra_scheduler1 == "##MOCK --exclusive=user\n##MOCK --time=2"
    extra_env_vars0 = s.extra_env_vars(index=0)
    extra_env_vars1 = s.extra_env_vars(index=1)
    assert extra_env_vars0.startswith("export from=func\n")
    assert extra_env_vars1.startswith("export from=static\n")
    extra_script0 = s.extra_script(index=0)
    extra_script1 = s.extra_script(index=1)
    assert extra_script0 == "echo 'func'"
    assert extra_script1 == "echo 'static'"
    cores0 = s._get_cores(index=0)
    cores1 = s._get_cores(index=1)
    assert cores0 == 4
    assert cores1 == 2
    executor_type0 = s._get_executor_type(index=0)
    executor_type1 = s._get_executor_type(index=1)
    assert executor_type0 == "ipyparallel"
    assert executor_type1 == "mpi4py"



"""Imports for the server_support module."""

from __future__ import annotations

from ._server_support.base_manager import BaseManager
from ._server_support.common import (
    _delete_old_ipython_profiles,
    _get_all_files,
    cleanup_scheduler_files,
    console,
    get_allowed_url,
    log,
    periodically_clean_ipython_profiles,
)
from ._server_support.database_manager import DatabaseManager
from ._server_support.job_manager import JobManager, MaxRestartsReachedError
from ._server_support.kill_manager import KillManager, logs_with_string_or_condition
from ._server_support.multi_run_manager import MultiRunManager
from ._server_support.parse_logs import _get_infos, parse_log_files
from ._server_support.run_manager import (
    RunManager,
    _start_after,
    _wait_for_finished,
    start_one_by_one,
)
from ._server_support.slurm_run import slurm_run

__all__ = [
    "DatabaseManager",
    "JobManager",
    "KillManager",
    "_wait_for_finished",
    "_start_after",
    "start_one_by_one",
    "logs_with_string_or_condition",
    "RunManager",
    "MultiRunManager",
    "slurm_run",
    "_get_infos",
    "parse_log_files",
    "BaseManager",
    "log",
    "console",
    "MaxRestartsReachedError",
    "get_allowed_url",
    "_get_all_files",
    "cleanup_scheduler_files",
    "_delete_old_ipython_profiles",
    "periodically_clean_ipython_profiles",
]



# Copyright (c) Microsoft Corporation. All rights reserved.
from __future__ import annotations

from pathlib import Path

import versioningit

PROJECT_DIR = Path(__file__).parent.parent
__version__ = "2.14.2"



"""Test log generation functions."""

from __future__ import annotations

import asyncio
import json
import logging

import adaptive
import pytest

from adaptive_scheduler import client_support
from adaptive_scheduler.utils import _at_least_adaptive_version


def expected_log_keys(learner: adaptive.BaseLearner) -> list[str]:
    """Return the expected keys for the log entry."""
    # Check if the result contains the expected keys
    expected_keys = [
        "elapsed_time",
        "overhead",
        "npoints",
        "cpu_usage",
        "mem_usage",
    ]
    if not _at_least_adaptive_version("0.16.0", raises=False) and not isinstance(
        learner,
        adaptive.SequenceLearner,
    ):
        # The loss cache for SequenceLearner was introduced in adaptive 0.16.0
        # see https://github.com/python-adaptive/adaptive/pull/411
        expected_keys.append("latest_loss")
    return expected_keys


@pytest.mark.asyncio()
async def test_get_log_entry(
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
) -> None:
    """Test `client_support._get_log_entry`."""
    # Prepare the runner and the learner
    learner = learners[0]
    runner = adaptive.Runner(
        learner,
        npoints_goal=1000,
        executor=adaptive.runner.SequentialExecutor(),
    )
    min_points = 100
    while learner.npoints < min_points:
        learner.loss()  # populate cache
        await asyncio.sleep(0.05)
    result = client_support._get_log_entry(runner, 0)

    # Check if the result contains the expected keys
    expected_keys = expected_log_keys(learner)
    assert all(key in result for key in expected_keys)


@pytest.mark.asyncio()
async def test_log_info(
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test `client_support.log_info`."""
    # Prepare the runner and the learner
    learner = learners[0]
    runner = adaptive.Runner(learner, shutdown_executor=True)
    learner.loss()  # populates loss cache and therefore latest_loss

    caplog.set_level(logging.INFO)
    interval = 0.05
    log_task = client_support.log_info(runner, interval)

    # Wait for some time to let the logging happen
    for _ in range(100):
        if log_task.done():
            assert log_task.exception() is None
        await asyncio.sleep(0.1)
        if len(caplog.records) > 10:
            break

    # Filter the captured log records based on level and logger name
    filtered_records = [
        r
        for r in caplog.records
        if r.levelno == logging.INFO and r.name == "adaptive_scheduler.client"
    ]

    # Check if the captured log records contain expected log entries
    assert len(filtered_records) > 0
    current_status_entries = 0
    for record in filtered_records:
        msg = record.getMessage()
        log_entry = json.loads(msg)
        if "current status" in log_entry["event"]:
            current_status_entries += 1
            # Check if the result contains the expected keys
            expected_keys = expected_log_keys(learner)
            assert all(key in log_entry for key in expected_keys)

    # Check if there were any "current status" log entries
    assert current_status_entries > 0, filtered_records

    runner.cancel()



"""Tests for the KillManager module."""

from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING
from unittest.mock import MagicMock

import pytest

from adaptive_scheduler.server_support import (
    DatabaseManager,
    KillManager,
    logs_with_string_or_condition,
)

if TYPE_CHECKING:
    from pathlib import Path


INTERVAL = 0.05


@pytest.fixture()
def kill_manager(db_manager: DatabaseManager) -> KillManager:
    """Fixture for creating a KillManager instance."""
    return KillManager(
        scheduler=db_manager.scheduler,
        database_manager=db_manager,
        error="srun: error:",
        interval=INTERVAL,
        max_cancel_tries=1,
        move_to=None,
    )


@pytest.mark.asyncio()
async def test_kill_manager_init(kill_manager: KillManager) -> None:
    """Test KillManager initialization."""
    assert kill_manager.scheduler is not None
    assert kill_manager.database_manager is not None
    assert kill_manager.error == "srun: error:"
    assert kill_manager.interval == INTERVAL
    assert kill_manager.max_cancel_tries == 1
    assert kill_manager.move_to is None
    assert kill_manager.cancelled == []
    assert kill_manager.deleted == []


def test_logs_with_string_or_condition_invalid_error() -> None:
    """Test logs_with_string_or_condition with invalid error type."""
    database_manager = MagicMock(spec=DatabaseManager)
    with pytest.raises(TypeError, match="`error` can only be a `str` or `callable`."):
        logs_with_string_or_condition(123, database_manager)  # type: ignore[arg-type]


def test_logs_with_string_or_condition_string_error(tmp_path: Path) -> None:
    """Test logs_with_string_or_condition with string error."""
    database_manager = MagicMock(spec=DatabaseManager)
    logs_file = tmp_path / "logs.txt"
    logs_file.write_text("Error: Something went wrong.")

    database_manager.as_dicts.return_value = [
        {
            "job_id": 1,
            "job_name": "test_job",
            "output_logs": [str(logs_file)],
            "log_fname": "log_file.log",
        },
    ]

    error = "Something went wrong"
    result = logs_with_string_or_condition(error, database_manager)
    assert len(result) == 1
    assert result[0][0] == "test_job"
    assert result[0][1][0] == str(logs_file)


def test_logs_with_string_or_condition_callable_error(tmp_path: Path) -> None:
    """Test logs_with_string_or_condition with callable error."""
    database_manager = MagicMock(spec=DatabaseManager)
    logs_file = tmp_path / "logs.txt"
    logs_file.write_text("Error: Something went wrong.")

    database_manager.as_dicts.return_value = [
        {
            "job_id": 1,
            "job_name": "test_job",
            "output_logs": [str(logs_file)],
            "log_fname": "log_file.log",
        },
    ]

    def custom_error(lines: list[str]) -> bool:
        return "Error" in "".join(lines)

    result = logs_with_string_or_condition(custom_error, database_manager)
    assert len(result) == 1
    assert result[0][0] == "test_job"
    assert result[0][1][0] == str(logs_file)


def test_logs_with_string_or_condition_no_error(tmp_path: Path) -> None:
    """Test logs_with_string_or_condition with no error."""
    database_manager = MagicMock(spec=DatabaseManager)
    logs_file = tmp_path / "logs.txt"
    logs_file.write_text("Info: Everything is fine.")

    database_manager.as_dicts.return_value = [
        {
            "job_id": 1,
            "job_name": "test_job",
            "output_logs": [str(logs_file)],
            "log_fname": "log_file.log",
        },
    ]

    error = "Something went wrong"
    result = logs_with_string_or_condition(error, database_manager)
    assert len(result) == 0


def test_logs_with_string_or_condition_missing_file() -> None:
    """Test logs_with_string_or_condition with missing file."""
    database_manager = MagicMock(spec=DatabaseManager)

    database_manager.as_dicts.return_value = [
        {
            "job_id": 1,
            "job_name": "test_job",
            "output_logs": ["non_existent_file.txt"],
            "log_fname": "log_file.log",
        },
    ]

    error = "Something went wrong"
    result = logs_with_string_or_condition(error, database_manager)
    assert len(result) == 0


@pytest.mark.asyncio()
async def test_kill_manager_manage(kill_manager: KillManager) -> None:
    """Test KillManager.manage method."""
    # The KillManager will read from the .out files, which are determined
    # from the scheduler.
    output_file_path = kill_manager.database_manager._output_logs("0", "test_job")[0]
    with output_file_path.open("w") as f:
        f.write("srun: error: GPU on fire!\n")

    kill_manager.database_manager.start()  # creates empty db
    kill_manager.start()
    assert kill_manager.task is not None
    # Marks the job as running, and sets the job_id, job_name, and log_fname
    index, _ = kill_manager.database_manager._choose_fname()
    kill_manager.database_manager._confirm_submitted(index, "test_job")
    kill_manager.database_manager._start_request("0", "log_fname.log", "test_job")
    kill_manager.scheduler.start_job("test_job")
    await asyncio.sleep(0.1)  # Give it some time to start and cancel the job
    assert "test_job" in kill_manager.cancelled
    assert str(output_file_path) in kill_manager.deleted


@pytest.mark.asyncio()
async def test_kill_manager_manage_exception(
    kill_manager: KillManager,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test KillManager.manage method with an exception."""
    # This will cause a ValueError when calling `_manage`
    kill_manager.error = 12345  # type: ignore[assignment]
    kill_manager.database_manager.start()  # creates empty db
    kill_manager.start()
    assert kill_manager.task is not None
    await asyncio.sleep(0.2)  # Give it some time to start and raise the exception

    timeout = 0.1
    try:
        await asyncio.wait_for(kill_manager.task, timeout=timeout)
    except asyncio.TimeoutError:
        kill_manager.task.cancel()

    with pytest.raises(asyncio.CancelledError):
        await kill_manager.task

    assert "got exception in kill manager" in caplog.text
    assert "TypeError" in caplog.text


@pytest.mark.asyncio()
async def test_kill_manager_manage_canceled(
    kill_manager: KillManager,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test KillManager.manage method when canceled."""
    caplog.set_level(logging.INFO)
    kill_manager.error = "never going to happen"
    kill_manager.database_manager.start()  # creates empty db
    kill_manager.start()
    assert kill_manager.task is not None
    await asyncio.sleep(0.1)  # Give it some time to start and raise the exception
    kill_manager.task.cancel()

    with pytest.raises(asyncio.CancelledError):
        await kill_manager.task

    assert "task was cancelled because of a CancelledError" in caplog.text



"""The DatabaseManager."""

from __future__ import annotations

import json
import pickle
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any

import pandas as pd
import zmq
import zmq.asyncio
import zmq.ssh

from adaptive_scheduler.utils import (
    _deserialize,
    _now,
    _serialize,
    cloudpickle_learners,
)

from .base_manager import BaseManager
from .common import log

if TYPE_CHECKING:
    from collections.abc import Callable

    import adaptive

    from adaptive_scheduler.scheduler import BaseScheduler


ctx = zmq.asyncio.Context()
FnameType = str | Path | list[str] | list[Path]
FnamesTypes = list[str] | list[Path] | list[list[str]] | list[list[Path]]


class JobIDExistsInDbError(Exception):
    """Raised when a job id already exists in the database."""


def _ensure_str(
    fnames: str | Path | FnamesTypes,
) -> str | list[str] | list[list[str]]:
    """Make sure that `pathlib.Path`s are converted to strings."""
    if isinstance(fnames, str | Path):
        return str(fnames)

    if isinstance(fnames, list | tuple):
        if len(fnames) == 0:
            return []  # type: ignore[return-value]
        if isinstance(fnames[0], str | Path):
            return [str(f) for f in fnames]
        if isinstance(fnames[0], list):
            return [[str(f) for f in sublist] for sublist in fnames]  # type: ignore[union-attr]
    msg = (
        "Invalid input: expected a  string/Path, or list of"
        " strings/Paths, a list of lists of strings/Paths."
    )
    raise ValueError(msg)


@dataclass
class _DBEntry:
    fname: str | list[str]
    job_id: str | None = None
    is_pending: bool = False
    is_done: bool = False
    log_fname: str | None = None
    job_name: str | None = None
    output_logs: list[str] = field(default_factory=list)
    start_time: float | None = None
    depends_on: list[int] = field(default_factory=list)


class SimpleDatabase:
    def __init__(self, db_fname: str | Path, *, clear_existing: bool = False) -> None:
        self.db_fname = Path(db_fname)
        self._data: list[_DBEntry] = []
        self._meta: dict[str, Any] = {}

        if self.db_fname.exists():
            if clear_existing:
                self.db_fname.unlink()
            else:
                with self.db_fname.open() as f:
                    raw_data = json.load(f)
                    self._data = [_DBEntry(**entry) for entry in raw_data["data"]]

    def all(self) -> list[_DBEntry]:
        return self._data

    def insert_multiple(self, entries: list[_DBEntry]) -> None:
        self._data.extend(entries)
        self._save()

    def update(self, update_dict: dict, indices: list[int] | None = None) -> None:
        for index, entry in enumerate(self._data):
            if indices is None or index in indices:
                for key, value in update_dict.items():
                    assert hasattr(entry, key)
                    setattr(entry, key, value)
        self._save()

    def count(self, condition: Callable[[_DBEntry], bool]) -> int:
        return sum(1 for entry in self._data if condition(entry))

    def get_with_index(self, condition: Callable[[_DBEntry], bool]) -> tuple[int, _DBEntry] | None:
        for index, entry in enumerate(self._data):
            if condition(entry):
                return index, entry
        return None

    def get(self, condition: Callable[[_DBEntry], bool]) -> _DBEntry | None:
        index_entry = self.get_with_index(condition)
        if index_entry is None:
            return None
        _, entry = index_entry
        return entry

    def get_all(
        self,
        condition: Callable[[_DBEntry], bool],
    ) -> list[tuple[int, _DBEntry]]:
        return [(i, entry) for i, entry in enumerate(self._data) if condition(entry)]

    def contains(self, condition: Callable[[_DBEntry], bool]) -> bool:
        return any(condition(entry) for entry in self._data)

    def as_dicts(self) -> list[dict[str, Any]]:
        return [asdict(entry) for entry in self._data]

    def _save(self) -> None:
        with self.db_fname.open("w") as f:
            json.dump({"data": self.as_dicts(), "meta": self._meta}, f, indent=4)

    def dependencies_satisfied(self, entry: _DBEntry) -> bool:
        return all(self._data[i].is_done for i in entry.depends_on)


class DatabaseManager(BaseManager):
    """Database manager.

    Parameters
    ----------
    url
        The url of the database manager, with the format
        ``tcp://ip_of_this_machine:allowed_port.``. Use `get_allowed_url`
        to get a `url` that will work.
    scheduler
        A scheduler instance from `adaptive_scheduler.scheduler`.
    db_fname
        Filename of the database, e.g. 'running.json'.
    learners
        List of `learners` corresponding to `fnames`.
    fnames
        List of `fnames` corresponding to `learners`.
    dependencies
        Dictionary of dependencies, e.g., ``{1: [0]}`` means that the ``learners[1]``
        depends on the ``learners[0]``. This means that the ``learners[1]`` will only
        start when the ``learners[0]`` is done.
    overwrite_db
        Overwrite the existing database upon starting.
    initializers
        List of functions that are called before the job starts, can populate
        a cache.

    Attributes
    ----------
    failed : list
        A list of entries that have failed and have been removed from the database.

    """

    def __init__(
        self,
        url: str,
        scheduler: BaseScheduler,
        db_fname: str | Path,
        learners: list[adaptive.BaseLearner],
        fnames: FnamesTypes,
        *,
        dependencies: dict[int, list[int]] | None = None,
        overwrite_db: bool = True,
        initializers: list[Callable[[], None]] | None = None,
        with_progress_bar: bool = True,
    ) -> None:
        super().__init__()
        self.url = url
        self.scheduler = scheduler
        self.db_fname = Path(db_fname)
        self.learners = learners
        self.fnames = fnames
        self.dependencies = dependencies or {}
        self.overwrite_db = overwrite_db
        self.initializers = initializers
        self.with_progress_bar = with_progress_bar
        self._last_reply: str | list[str] | Exception | None = None
        self._last_request: tuple[str, ...] | None = None
        self.failed: list[dict[str, Any]] = []
        self._pickling_time: float | None = None
        self._total_learner_size: int | None = None
        self._db: SimpleDatabase | None = None

    def _setup(self) -> None:
        if self.db_fname.exists() and not self.overwrite_db:
            return
        self.create_empty_db()
        self._total_learner_size, self._pickling_time = cloudpickle_learners(
            self.learners,
            self.fnames,
            initializers=self.initializers,
            with_progress_bar=self.with_progress_bar,
        )

    def update(self, queue: dict[str, dict[str, str]] | None = None) -> None:
        """If the ``job_id`` isn't running anymore, replace it with None."""
        if self._db is None:
            return
        if queue is None:
            queue = self.scheduler.queue(me_only=True)
        job_names_in_queue = [x["job_name"] for x in queue.values()]
        failed = self._db.get_all(
            lambda e: e.job_name is not None and e.job_name not in job_names_in_queue,  # type: ignore[operator]
        )
        self.failed.extend([asdict(entry) for _, entry in failed])
        indices = [index for index, _ in failed]
        self._db.update(
            {"job_id": None, "job_name": None, "is_pending": False},
            indices,
        )

    def n_done(self) -> int:
        """Return the number of jobs that are done."""
        if self._db is None:
            return 0
        return self._db.count(lambda e: e.is_done)

    def n_unscheduled(self) -> int:
        """Return the number of jobs that are not scheduled."""
        if self._db is None:
            return 0
        return self._db.count(lambda e: not e.is_done and not e.is_pending)

    def is_done(self) -> bool:
        """Return True if all jobs are done."""
        return self.n_done() == len(self.fnames)

    def create_empty_db(self) -> None:
        """Create an empty database.

        It keeps track of ``fname -> (job_id, is_done, log_fname, job_name)``.
        """
        deps = self.dependencies
        entries: list[_DBEntry] = [
            _DBEntry(fname=fname, depends_on=deps.get(i, []))  # type: ignore[arg-type]
            for i, fname in enumerate(_ensure_str(self.fnames))
        ]
        if self.db_fname.exists():
            self.db_fname.unlink()
        self._db = SimpleDatabase(self.db_fname)
        self._db.insert_multiple(entries)

    def as_dicts(self) -> list[dict[str, str]]:
        """Return the database as a list of dictionaries."""
        if self._db is None:
            return []
        return self._db.as_dicts()

    def as_df(self) -> pd.DataFrame:
        """Return the database as a `pandas.DataFrame`."""
        return pd.DataFrame(self.as_dicts())

    def _output_logs(self, job_id: str, job_name: str) -> list[Path]:
        job_id = self.scheduler.sanatize_job_id(job_id)
        output_fnames = self.scheduler.output_fnames(job_name)
        return [
            f.with_name(f.name.replace(self.scheduler._JOB_ID_VARIABLE, job_id))
            for f in output_fnames
        ]

    def _choose_fname(self) -> tuple[int, str | list[str] | None]:
        assert self._db is not None
        entry = self._db.get(
            lambda e: e.job_id is None
            and not e.is_done
            and not e.is_pending
            and self._db.dependencies_satisfied(e),  # type: ignore[union-attr]
        )
        if all(e.is_done for e in self._db.all()):
            msg = "Requested a new job but no more learners to run in the database."
            raise RuntimeError(msg)
        if entry is None:
            # Currently, we cannot schedule any more jobs, because we're waiting
            # for dependencies to be satisfied.
            return -1, None
        log.debug("choose fname", entry=entry)
        index = self._db.all().index(entry)
        return index, _ensure_str(entry.fname)  # type: ignore[return-value]

    def _confirm_submitted(self, index: int, job_name: str) -> None:
        assert self._db is not None
        self._db.update(
            {
                "job_name": job_name,
                "is_pending": True,
            },
            indices=[index],
        )

    def _start_request(
        self,
        job_id: str,
        log_fname: str,
        job_name: str,
    ) -> str | list[str] | None:
        assert self._db is not None
        if self._db.contains(lambda e: e.job_id == job_id):
            entry = self._db.get(lambda e: e.job_id == job_id)
            assert entry is not None
            fname = entry.fname  # already running
            msg = (
                f"The job_id {job_id} already exists in the database and "
                f"runs {fname}. You might have forgotten to use the "
                "`if __name__ == '__main__': ...` idiom in your code. Read the "
                "warning in the [mpi4py](https://bit.ly/2HAk0GG) documentation.",
            )
            raise JobIDExistsInDbError(msg)
        entry = self._db.get(lambda e: e.job_name == job_name and e.is_pending)
        log.debug("choose fname", entry=entry)
        if entry is None:
            return None
        index = self._db.all().index(entry)
        self._db.update(
            {
                "job_id": job_id,
                "log_fname": log_fname,
                "output_logs": _ensure_str(self._output_logs(job_id, job_name)),
                "start_time": _now(),
                "is_pending": False,
            },
            indices=[index],
        )
        return _ensure_str(entry.fname)  # type: ignore[return-value]

    def _stop_request(self, fname: str | list[str] | Path | list[Path]) -> None:
        fname_str = _ensure_str(fname)
        reset = {"job_id": None, "is_done": True, "job_name": None, "is_pending": False}
        assert self._db is not None
        entry_indices = [index for index, _ in self._db.get_all(lambda e: e.fname == fname_str)]
        self._db.update(reset, entry_indices)

    def _stop_requests(self, fnames: FnamesTypes) -> None:
        # Same as `_stop_request` but optimized for processing many `fnames` at once
        assert self._db is not None
        fnames_str = {str(fname) for fname in _ensure_str(fnames)}
        reset = {"job_id": None, "is_done": True, "job_name": None, "is_pending": False}
        entry_indices = [
            index for index, _ in self._db.get_all(lambda e: str(e.fname) in fnames_str)
        ]
        self._db.update(reset, entry_indices)

    def _dispatch(
        self,
        request: tuple[str, str | list[str]] | tuple[str],
    ) -> str | list[str] | Exception | None:
        request_type, *request_arg = request
        log.debug("got a request", request=request)
        try:
            if request_type == "start":
                # workers send us their slurm ID for us to fill in
                job_id, log_fname, job_name = request_arg
                # give the worker a job and send back the fname to the worker
                fname = self._start_request(job_id, log_fname, job_name)  # type: ignore[arg-type]
                if fname is None:
                    # This should never happen because the _manage co-routine
                    # should have stopped the workers before this happens.
                    msg = "No more learners to run in the database."
                    raise RuntimeError(msg)  # noqa: TRY301
                log.debug(
                    "choose a fname",
                    fname=fname,
                    job_id=job_id,
                    log_fname=log_fname,
                    job_name=job_name,
                )
                return fname
            if request_type == "stop":
                fname = request_arg[0]  # workers send us the fname they were given
                log.debug("got a stop request", fname=fname)
                self._stop_request(fname)  # reset the job_id to None
                return None
        except Exception as e:  # noqa: BLE001
            return e
        msg = f"Unknown request type: {request_type}"
        raise ValueError(msg)

    async def _manage(self) -> None:
        """Database manager co-routine.

        Returns
        -------
        coroutine

        """
        log.debug("started database")
        socket = ctx.socket(zmq.REP)
        socket.bind(self.url)
        try:
            while True:
                try:
                    self._last_request = await socket.recv_serialized(_deserialize)
                except zmq.error.Again:
                    log.exception(
                        "socket.recv_serialized failed in the DatabaseManager"
                        " with `zmq.error.Again`.",
                    )
                except pickle.UnpicklingError as e:
                    if r"\x03" in str(e):
                        # Empty frame received.
                        # TODO: not sure why this happens
                        pass
                    else:
                        log.exception(
                            "socket.recv_serialized failed in the DatabaseManager"
                            " with `pickle.UnpicklingError` in _deserialize.",
                        )
                else:
                    assert self._last_request is not None  # for mypy
                    self._last_reply = self._dispatch(self._last_request)  # type: ignore[arg-type]
                    await socket.send_serialized(self._last_reply, _serialize)
                if self.is_done():
                    break
        finally:
            socket.close()

    def replace_learner(self, index: int, new_learner: adaptive.BaseLearner) -> None:
        """Replace a learner and update the corresponding database entry and cloudpickled file.

        Parameters
        ----------
        index
            The index of the learner to replace.
        new_learner
            The new learner to replace the old one.

        """
        if index < 0 or index >= len(self.learners):
            msg = "Index out of range"
            raise IndexError(msg)

        fname = self.fnames[index]

        # Update the database entry
        assert self._db is not None
        index_entry = self._db.get_with_index(lambda e: e.fname == _ensure_str(fname))
        assert index_entry is not None
        index, entry = index_entry
        if entry.is_done:
            msg = f"Learner at index {index} is already done and cannot be replaced."
            raise ValueError(msg)
        assert not entry.is_pending
        assert entry.job_id is None

        # Replace the learner in the list
        self.learners[index] = new_learner

        # Cloudpickle the new learner
        cloudpickle_learners(
            [new_learner],
            [fname],  # type: ignore[arg-type]
        )
        # Note that self._total_learner_size and self._pickling_time are
        # not updated now! But we don't care about that.

        log.debug(f"Replaced learner at index {index} with a new learner")



"""Tests for the `client_support` module."""

from __future__ import annotations

import tempfile
from unittest import mock

import adaptive
import pytest
import zmq

from adaptive_scheduler import client_support


@pytest.fixture(scope="module")
def zmq_url() -> str:
    """Create a ZMQ URL for testing purposes."""
    return "tcp://127.0.0.1:5555"


@pytest.fixture(scope="module")
def client(zmq_url: str) -> zmq.Socket:
    """Create a client for testing purposes."""
    ctx = zmq.Context()
    client = ctx.socket(zmq.REQ)
    client.connect(zmq_url)
    return client


@pytest.mark.asyncio()
async def test_get_learner(zmq_url: str) -> None:
    """Test `get_learner` function."""
    with tempfile.NamedTemporaryFile() as tmpfile:
        log_fname = tmpfile.name
        job_id = "test_job_id"
        job_name = "test_job_name"

        # Mock the fname_to_learner function to avoid creating real learners

        with mock.patch(
            "adaptive_scheduler.client_support.fname_to_learner",
        ) as mock_fname_to_learner:
            learner = adaptive.Learner1D(lambda x: x, (0, 1))
            initializer = None
            mock_fname_to_learner.return_value = (learner, initializer)
            # Patch `socket.recv_serialized` to return the desired reply
            with mock.patch(
                "zmq.sugar.socket.Socket.recv_serialized",
                return_value="Simple reply",
            ):
                learner, fname, initializer = client_support.get_learner(
                    zmq_url,
                    log_fname,
                    job_id,
                    job_name,
                )
                assert isinstance(learner, adaptive.Learner1D)
                assert isinstance(fname, str)
                assert initializer is None

            # Test no more learners available
            with (
                mock.patch(
                    "zmq.sugar.socket.Socket.recv_serialized",
                    return_value=None,
                ),
                pytest.raises(RuntimeError, match="No learners to be run"),
            ):
                client_support.get_learner(
                    zmq_url,
                    log_fname,
                    job_id,
                    job_name,
                )

            # Test return exception
            with (
                mock.patch("adaptive_scheduler.client_support.log") as mock_log,
                mock.patch(
                    "zmq.sugar.socket.Socket.recv_serialized",
                    return_value=ValueError("Yo"),
                ),
            ):
                with pytest.raises(
                    ValueError,
                    match="Yo",
                ):
                    client_support.get_learner(
                        zmq_url,
                        log_fname,
                        job_id,
                        job_name,
                    )
                mock_log.exception.assert_called_with("got an exception")


@pytest.mark.asyncio()
async def test_tell_done(zmq_url: str) -> None:
    """Test `tell_done` function."""
    fname = "test_learner_file.pkl"
    with (
        mock.patch("adaptive_scheduler.client_support.log") as mock_log,
        mock.patch("zmq.sugar.socket.Socket.recv_serialized", return_value=None),
    ):
        client_support.tell_done(zmq_url, fname)
        mock_log.info.assert_called_with(
            "sent stop signal, going to wait 300s for a reply",
            fname="test_learner_file.pkl",
        )



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any

from adaptive_scheduler.scheduler import SLURM, slurm_partitions
from adaptive_scheduler.utils import _get_default_args

from .common import console
from .run_manager import RunManager

if TYPE_CHECKING:
    from collections.abc import Callable

    import adaptive

    from adaptive_scheduler.utils import (
        _DATAFRAME_FORMATS,
        EXECUTOR_TYPES,
        LOKY_START_METHODS,
        GoalTypes,
    )


def slurm_run(
    learners: list[adaptive.BaseLearner],
    fnames: list[str] | list[Path],
    *,
    # Specific to slurm_run
    name: str = "adaptive-scheduler",
    folder: str | Path = "",
    # SLURM scheduler arguments
    partition: str | tuple[str | Callable[[], str], ...] | None = None,
    nodes: int | tuple[int | None | Callable[[], int | None], ...] | None = 1,
    cores_per_node: int | tuple[int | None | Callable[[], int | None], ...] | None = None,
    num_threads: int | tuple[int | Callable[[], int], ...] = 1,
    exclusive: bool | tuple[bool | Callable[[], bool], ...] = False,
    executor_type: EXECUTOR_TYPES
    | tuple[EXECUTOR_TYPES | Callable[[], EXECUTOR_TYPES], ...] = "process-pool",
    extra_scheduler: list[str] | tuple[list[str] | Callable[[], list[str]], ...] | None = None,
    # Same as RunManager below (except job_name, move_old_logs_to, and db_fname)
    goal: GoalTypes | None = None,
    check_goal_on_start: bool = True,
    dependencies: dict[int, list[int]] | None = None,
    runner_kwargs: dict | None = None,
    url: str | None = None,
    save_interval: float = 300,
    log_interval: float = 300,
    job_manager_interval: float = 60,
    kill_interval: float = 60,
    kill_on_error: str | Callable[[list[str]], bool] | None = "srun: error:",
    overwrite_db: bool = True,
    job_manager_kwargs: dict[str, Any] | None = None,
    kill_manager_kwargs: dict[str, Any] | None = None,
    loky_start_method: LOKY_START_METHODS = "loky",
    cleanup_first: bool = True,
    save_dataframe: bool = True,
    dataframe_format: _DATAFRAME_FORMATS = "pickle",
    max_log_lines: int = 500,
    max_fails_per_job: int = 50,
    max_simultaneous_jobs: int = 100,
    initializers: list[Callable[[], None]] | None = None,
    quiet: bool = False,
    # RunManager arguments
    extra_run_manager_kwargs: dict[str, Any] | None = None,
    extra_scheduler_kwargs: dict[str, Any] | None = None,
) -> RunManager:
    """Run adaptive on a SLURM cluster.

    ``cores_per_node``, ``nodes``, ``extra_scheduler``,
    ``executor_type``, ``exclusive``,
    ``num_threads`` and ``partition`` can be either a single value or a tuple of
    values. If a tuple is given, then the length of the tuple should be the same
    as the number of learners (jobs) that are run. This allows for different
    resources for different jobs. The tuple elements are also allowed to be
    callables without arguments, which will be called when the job is submitted.
    These callables should return the value that is needed. See the type hints
    for the allowed types.

    Parameters
    ----------
    learners
        A list of learners.
    fnames
        A list of filenames to save the learners.
    name
        The name of the job.
    folder
        The folder to save the adaptive_scheduler files such as logs, database,
        and ``.sbatch`` files in.
    partition
        The partition to use. If None, then the default partition will be used.
        (The one marked with a * in `sinfo`). Use
        `adaptive_scheduler.scheduler.slurm_partitions` to see the
        available partitions.
    nodes
        The number of nodes to use.
    cores_per_node
        The number of cores per node to use. If None, then all cores on the partition
        will be used.
    num_threads
        The number of threads to use.
    exclusive
        Whether to use exclusive nodes, adds ``"--exclusive"`` if True.
    executor_type
        The executor that is used, by default `concurrent.futures.ProcessPoolExecutor` is used.
        One can use ``"ipyparallel"``, ``"dask-mpi"``, ``"mpi4py"``,
        ``"loky"``, ``"sequential"``, or ``"process-pool"``.
    extra_scheduler
        Extra ``#SLURM`` (depending on scheduler type)
        arguments, e.g. ``["--exclusive=user", "--time=1"]`` or a tuple of lists,
        e.g. ``(["--time=10"], ["--time=20"]])`` for two jobs.
    goal
        The goal passed to the `adaptive.Runner`. Note that this function will
        be serialized and pasted in the ``job_script``. Can be a smart-goal
        that accepts
        ``Callable[[adaptive.BaseLearner], bool] | float | datetime | timedelta | None``.
        See `adaptive_scheduler.utils.smart_goal` for more information.
    check_goal_on_start
        Checks whether a learner is already done. Only works if the learner is loaded.
    dependencies
        Dictionary of dependencies, e.g., ``{1: [0]}`` means that the ``learners[1]``
        depends on the ``learners[0]``. This means that the ``learners[1]`` will only
        start when the ``learners[0]`` is done.
    runner_kwargs
        Extra keyword argument to pass to the `adaptive.Runner`. Note that this dict
        will be serialized and pasted in the ``job_script``.
    url
        The url of the database manager, with the format
        ``tcp://ip_of_this_machine:allowed_port.``. If None, a correct url will be chosen.
    save_interval
        Time in seconds between saving of the learners.
    log_interval
        Time in seconds between log entries.
    job_manager_interval
        Time in seconds between checking and starting jobs.
    kill_interval
        Check for `kill_on_error` string inside the log-files every `kill_interval` seconds.
    kill_on_error
        If ``error`` is a string and is found in the log files, the job will
        be cancelled and restarted. If it is a callable, it is applied
        to the log text. Must take a single argument, a list of
        strings, and return True if the job has to be killed, or
        False if not. Set to None if no `KillManager` is needed.
    overwrite_db
        Overwrite the existing database.
    job_manager_kwargs
        Keyword arguments for the `JobManager` function that aren't set in ``__init__`` here.
    kill_manager_kwargs
        Keyword arguments for the `KillManager` function that aren't set in ``__init__`` here.
    loky_start_method
        Loky start method, by default "loky".
    cleanup_first
        Cancel all previous jobs generated by the same RunManager and clean logfiles.
    save_dataframe
        Whether to periodically save the learner's data as a `pandas.DataFame`.
    dataframe_format
        The format in which to save the `pandas.DataFame`. See the type hint for the options.
    max_log_lines
        The maximum number of lines to display in the log viewer widget.
    max_fails_per_job
        Maximum number of times that a job can fail. This is here as a fail switch
        because a job might fail instantly because of a bug inside your code.
        The job manager will stop when
        ``n_jobs * total_number_of_jobs_failed > max_fails_per_job`` is true.
    max_simultaneous_jobs
        Maximum number of simultaneously running jobs. By default no more than 500
        jobs will be running. Keep in mind that if you do not specify a ``runner.goal``,
        jobs will run forever, resulting in the jobs that were not initially started
        (because of this `max_simultaneous_jobs` condition) to not ever start.
    initializers
        List of functions that are called before the job starts, can populate
        a cache.
    quiet
        Whether to show a progress bar when creating learner files.
    extra_run_manager_kwargs
        Extra keyword arguments to pass to the `RunManager`.
    extra_scheduler_kwargs
        Extra keyword arguments to pass to the `adaptive_scheduler.scheduler.SLURM`.

    Returns
    -------
    RunManager

    """
    if partition is None:
        partitions = slurm_partitions()
        assert isinstance(partitions, dict)
        partition, ncores = next(iter(partitions.items()))
        if not quiet:
            console.log(
                f"Using default partition {partition} (The one marked"
                f" with a '*' in `sinfo`) with {ncores} cores."
                " Use `adaptive_scheduler.scheduler.slurm_partitions`"
                " to see the available partitions.",
            )
    if (
        executor_type == "process-pool"
        and nodes is not None
        and (
            nodes > 1 if isinstance(nodes, int) else any(n > 1 for n in nodes if isinstance(n, int))
        )
    ):
        msg = "process-pool can maximally use a single node, use e.g., ipyparallel for multi node."
        raise ValueError(msg)
    folder = Path(folder)
    folder.mkdir(parents=True, exist_ok=True)
    if cores_per_node is None:
        if isinstance(partition, tuple) and any(callable(p) for p in partition):
            msg = "cores_per_node must be given if partition is a callable."
            raise ValueError(msg)
        partitions = slurm_partitions()
        assert isinstance(partitions, dict)
        cores_per_node = (
            tuple(partitions[p] for p in partition)  # type: ignore[misc,index]
            if isinstance(partition, tuple)
            else partitions[partition]
        )

    if extra_scheduler_kwargs is None:
        extra_scheduler_kwargs = {}
    if extra_scheduler is not None:
        # "extra_scheduler" used to be passed via the extra_scheduler_kwargs
        # this ensures backwards compatibility
        assert "extra_scheduler" not in extra_scheduler_kwargs
        extra_scheduler_kwargs["extra_scheduler"] = extra_scheduler

    slurm_kwargs = dict(
        _get_default_args(SLURM),
        nodes=nodes,
        cores_per_node=cores_per_node,
        partition=partition,
        log_folder=folder / "logs",
        batch_folder=folder / "batch_scripts",
        executor_type=executor_type,
        num_threads=num_threads,
        exclusive=exclusive,
        **extra_scheduler_kwargs,
    )
    scheduler = SLURM(**slurm_kwargs)
    # Below are the defaults for the RunManager
    kw = dict(
        _get_default_args(RunManager),
        goal=goal,
        check_goal_on_start=check_goal_on_start,
        dependencies=dependencies,
        runner_kwargs=runner_kwargs,
        url=url,
        save_interval=save_interval,
        log_interval=log_interval,
        job_name=name,
        job_manager_interval=job_manager_interval,
        kill_interval=kill_interval,
        kill_on_error=kill_on_error,
        move_old_logs_to=folder / "old_logs",
        db_fname=folder / f"{name}.db.json",
        overwrite_db=overwrite_db,
        job_manager_kwargs=job_manager_kwargs,
        kill_manager_kwargs=kill_manager_kwargs,
        loky_start_method=loky_start_method,
        cleanup_first=cleanup_first,
        save_dataframe=save_dataframe,
        dataframe_format=dataframe_format,
        max_log_lines=max_log_lines,
        max_fails_per_job=max_fails_per_job,
        max_simultaneous_jobs=max_simultaneous_jobs,
        initializers=initializers,
        quiet=quiet,
        scheduler=scheduler,
        learners=learners,
        fnames=fnames,
    )
    if extra_run_manager_kwargs is None:
        extra_run_manager_kwargs = {}
    return RunManager(**dict(kw, **extra_run_manager_kwargs))



"""BaseScheduler for Adaptive Scheduler."""

from __future__ import annotations

import abc
import subprocess
import sys
import textwrap
import time
import warnings
from pathlib import Path
from typing import TYPE_CHECKING

import pandas as pd

from adaptive_scheduler._scheduler.common import run_submit
from adaptive_scheduler.utils import EXECUTOR_TYPES, _progress

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any, ClassVar


_MULTI_LINE_BREAK = " \\\n    "


class BaseScheduler(abc.ABC):
    """Base object for a Scheduler.

    Parameters
    ----------
    cores
        Number of cores per job (so per learner.)
    python_executable
        The Python executable that should run adaptive-scheduler. By default
        it uses the same Python as where this function is called.
    log_folder
        The folder in which to put the log-files.
    mpiexec_executable
        ``mpiexec`` executable. By default `mpiexec` will be
        used (so probably from ``conda``).
    executor_type
        The executor that is used, by default `concurrent.futures.ProcessPoolExecutor` is used.
        One can use ``"ipyparallel"``, ``"dask-mpi"``, ``"mpi4py"``,
        ``"loky"``, ``"sequential"``, or ``"process-pool"``.
    num_threads
        ``MKL_NUM_THREADS``, ``OPENBLAS_NUM_THREADS``, ``OMP_NUM_THREADS``, and
        ``NUMEXPR_NUM_THREADS`` will be set to this number.
    extra_scheduler
        Extra ``#SLURM`` (depending on scheduler type)
        arguments, e.g. ``["--exclusive=user", "--time=1"]``.
    extra_env_vars
        Extra environment variables that are exported in the job
        script. e.g. ``["TMPDIR='/scratch'", "PYTHONPATH='my_dir:$PYTHONPATH'"]``.
    extra_script
        Extra script that will be executed after any environment variables are set,
        but before the main scheduler is run.
    batch_folder
        The folder in which to put the batch files.

    Returns
    -------
    `BaseScheduler` object.

    """

    _ext: ClassVar[str]
    _submit_cmd: ClassVar[str]
    _options_flag: ClassVar[str]
    _cancel_cmd: ClassVar[str]
    _JOB_ID_VARIABLE: ClassVar[str] = "${JOB_ID}"

    def __init__(
        self,
        cores: int | tuple[int | Callable[[], int], ...],
        *,
        python_executable: str | None = None,
        log_folder: str | Path = "",
        mpiexec_executable: str | None = None,
        executor_type: EXECUTOR_TYPES
        | tuple[EXECUTOR_TYPES | Callable[[], EXECUTOR_TYPES], ...] = "process-pool",
        num_threads: int | tuple[int | Callable[[], int], ...] = 1,
        extra_scheduler: list[str] | tuple[list[str] | Callable[[], list[str]], ...] | None = None,
        extra_env_vars: list[str] | tuple[list[str] | Callable[[], list[str]], ...] | None = None,
        extra_script: str | tuple[str | Callable[[], str], ...] | None = None,
        batch_folder: str | Path = "",
    ) -> None:
        """Initialize the scheduler."""
        self.cores = cores
        self.python_executable = python_executable or sys.executable
        self.log_folder = log_folder
        self.batch_folder = batch_folder
        self.mpiexec_executable = mpiexec_executable or "mpiexec"
        self.executor_type = executor_type
        self.num_threads = num_threads
        self._extra_scheduler = extra_scheduler
        self._extra_env_vars = extra_env_vars
        self._extra_script = extra_script if extra_script is not None else ""
        # This attribute is set in JobManager._setup ATM (hacky)
        self._command_line_options: dict[str, Any] | None = None

    @abc.abstractmethod
    def queue(self, *, me_only: bool = True) -> dict[str, dict]:
        """Get the current running and pending jobs.

        Parameters
        ----------
        me_only
            Only see your jobs.

        Returns
        -------
        queue
            Mapping of ``job_id`` -> `dict` with ``name`` and ``state``, for
            example ``{job_id: {"job_name": "TEST_JOB-1", "state": "R" or "Q"}}``.

        Notes
        -----
        This function might return extra information about the job, however
        this is not used elsewhere in this package.

        """

    def queue_df(self) -> pd.DataFrame:
        """Get the current running and pending jobs as a `pandas.DataFrame`."""
        queue = self.queue()
        return pd.DataFrame(queue).transpose()

    @property
    def ext(self) -> str:
        """The extension of the job script."""
        return self._ext

    @property
    def submit_cmd(self) -> str:
        """Command to start a job, e.g. ``qsub fname.batch`` or ``sbatch fname.sbatch``."""
        return self._submit_cmd

    @abc.abstractmethod
    def job_script(self, options: dict[str, Any], *, index: int | None = None) -> str:
        """Get a jobscript in string form.

        Returns
        -------
        job_script
            A job script that can be submitted to the scheduler.
        index
            The index of the job that is being run. This is used when
            specifying different resources for different jobs.

        """

    @property
    def single_job_script(self) -> bool:
        return isinstance(self.cores, int)

    def _get_executor_type(self, *, index: int | None = None) -> str:
        if self.single_job_script:
            assert isinstance(self.executor_type, str)
            return self.executor_type
        assert index is not None
        return _maybe_call(self.executor_type[index])

    def batch_fname(self, name: str) -> Path:
        """The filename of the job script."""
        if self.batch_folder:
            batch_folder = Path(self.batch_folder)
            batch_folder.mkdir(exist_ok=True, parents=True)
        else:
            batch_folder = Path.cwd()
        return batch_folder / f"{name}{self.ext}"

    @staticmethod
    def sanatize_job_id(job_id: str) -> str:
        """Sanatize the job_id."""
        return job_id

    def job_names_to_job_ids(self, *job_names: str) -> list[str]:
        """Get the job_ids from the job_names in the queue."""
        queue = self.queue()
        job_name_to_id = {info["job_name"]: job_id for job_id, info in queue.items()}
        return [job_name_to_id[job_name] for job_name in job_names if job_name in job_name_to_id]

    def cancel(
        self,
        job_names: list[str],
        *,
        with_progress_bar: bool = True,
        max_tries: int = 5,
    ) -> None:
        """Cancel all jobs in `job_names`.

        Parameters
        ----------
        job_names
            List of job names.
        with_progress_bar
            Display a progress bar using `tqdm`.
        max_tries
            Maximum number of attempts to cancel a job.

        """

        def cancel_jobs(job_ids: list[str]) -> None:
            for job_id in _progress(job_ids, with_progress_bar, "Canceling jobs"):
                cmd = f"{self._cancel_cmd} {job_id}".split()
                returncode = subprocess.run(
                    cmd,
                    stderr=subprocess.PIPE,
                    check=False,
                ).returncode
                if returncode != 0:
                    warnings.warn(
                        f"Couldn't cancel '{job_id}'.",
                        UserWarning,
                        stacklevel=2,
                    )

        job_names_set = set(job_names)
        for _ in range(max_tries):
            job_ids = self.job_names_to_job_ids(*job_names_set)
            if not job_ids:
                # no more running jobs
                break
            cancel_jobs(job_ids)
            time.sleep(0.5)

    def _expand_options(
        self,
        custom: tuple[str, ...],
        name: str,
        options: dict[str, Any],
    ) -> str:
        """Expand the options.

        This is used to expand the options that are passed to the job script.
        """
        log_fname = self.log_fname(name)
        return _MULTI_LINE_BREAK.join(
            (
                *custom,
                f"--log-fname {log_fname}",
                f"--job-id {self._JOB_ID_VARIABLE}",
                f"--name {name}",
                *(f"{k} {v}" if v is not None else k for k, v in options.items()),
            ),
        )

    def _get_cores(self, index: int | None = None) -> int:
        if self.single_job_script:
            cores = self.cores
        else:
            assert index is not None
            assert isinstance(self.cores, tuple)
            cores = _maybe_call(self.cores[index])
        assert isinstance(cores, int)
        return cores

    def _mpi4py(self, *, index: int | None = None) -> tuple[str, ...]:
        cores = self._get_cores(index=index)
        return (
            f"{self.mpiexec_executable}",
            f"-n {cores} {self.python_executable}",
            f"-m mpi4py.futures {self.launcher}",
        )

    def _dask_mpi(self, *, index: int | None = None) -> tuple[str, ...]:
        cores = self._get_cores(index=index)
        return (
            f"{self.mpiexec_executable}",
            f"-n {cores} {self.python_executable} {self.launcher}",
        )

    def _ipyparallel(self, *, index: int | None = None) -> tuple[str, tuple[str, ...]]:
        cores = self._get_cores(index=index)
        job_id = self._JOB_ID_VARIABLE
        profile = "${profile}"
        start = textwrap.dedent(
            f"""\
            profile=adaptive_scheduler_{job_id}

            echo "Creating profile {profile}"
            ipython profile create {profile}

            echo "Launching controller"
            ipcontroller --ip="*" --profile={profile} --log-to-file &
            sleep 10

            echo "Launching engines"
            {self.mpiexec_executable} \\
                -n {cores-1} \\
                ipengine \\
                --profile={profile} \\
                --mpi \\
                --cluster-id='' \\
                --log-to-file &

            echo "Starting the Python script"
            {self.python_executable} {self.launcher} \\
            """,
        )
        custom = (f"    --profile {profile}", f"--n {cores-1}")
        return start, custom

    def _process_pool(self) -> tuple[str, ...]:
        return (f"{self.python_executable} {self.launcher}",)

    def _sequential_executor(self) -> tuple[str, ...]:
        return (f"{self.python_executable} {self.launcher}",)

    def _executor_specific(
        self,
        name: str,
        options: dict[str, Any],
        *,
        index: int | None = None,
    ) -> str:
        start = ""
        executor_type = self._get_executor_type(index=index)
        if executor_type == "mpi4py":
            opts = self._mpi4py(index=index)
        elif executor_type == "dask-mpi":
            opts = self._dask_mpi(index=index)
        elif executor_type == "ipyparallel":
            cores = self._get_cores(index=index)
            if cores <= 1:
                msg = (
                    "`ipyparalllel` uses 1 cores of the `adaptive.Runner` and"
                    " the rest of the cores for the engines, so use more than 1 core."
                )
                raise ValueError(msg)
            start, opts = self._ipyparallel(index=index)
        elif executor_type in ("process-pool", "loky"):
            opts = self._process_pool()
        elif executor_type == "sequential":
            opts = self._sequential_executor()
        else:
            msg = (
                "Use 'ipyparallel', 'dask-mpi', 'mpi4py', 'loky', 'sequential', or 'process-pool'."
            )
            raise NotImplementedError(msg)
        return start + self._expand_options(opts, name, options)

    def log_fname(self, name: str) -> Path:
        """The filename of the log (with JOB_ID_VARIABLE)."""
        if self.log_folder:
            log_folder = Path(self.log_folder)
            log_folder.mkdir(exist_ok=True)
        else:
            log_folder = Path.cwd()
        return log_folder / f"{name}-{self._JOB_ID_VARIABLE}.log"

    def output_fnames(self, name: str) -> list[Path]:
        """Scheduler output file names (with JOB_ID_VARIABLE)."""
        log_fname = self.log_fname(name)
        return [log_fname.with_suffix(".out")]

    @property
    def launcher(self) -> Path:
        from adaptive_scheduler import _server_support

        return Path(_server_support.__file__).parent / "launcher.py"

    def _extra_scheduler_list(self, *, index: int | None = None) -> list[str]:
        if self._extra_scheduler is None:
            return []
        if self.single_job_script:
            assert isinstance(self._extra_scheduler, list)
            return self._extra_scheduler
        assert index is not None
        return _maybe_call(self._extra_scheduler[index])

    def extra_scheduler(self, *, index: int | None = None) -> str:
        """Scheduler options that go in the job script."""
        extra_scheduler: list[str] = self._extra_scheduler_list(index=index)
        return "\n".join(f"#{self._options_flag} {arg}" for arg in extra_scheduler)

    def _get_num_threads(self, *, index: int | None = None) -> int:
        if self.single_job_script:
            assert isinstance(self.num_threads, int)
            return self.num_threads
        assert index is not None
        return _maybe_call(self.num_threads[index])  # type: ignore[index]

    def extra_env_vars(self, *, index: int | None = None) -> str:
        """Environment variables that need to exist in the job script."""
        extra_env_vars: list[str]
        if self._extra_env_vars is None:
            extra_env_vars = []
        elif self.single_job_script:
            assert isinstance(self._extra_env_vars, list)
            extra_env_vars = self._extra_env_vars.copy()
        else:
            assert index is not None
            extra_env_vars = _maybe_call(self._extra_env_vars[index]).copy()  # type: ignore[union-attr]
        num_threads = self._get_num_threads(index=index)
        extra_env_vars.extend(
            [
                f"EXECUTOR_TYPE={self._get_executor_type(index=index)}",
                f"MKL_NUM_THREADS={num_threads}",
                f"OPENBLAS_NUM_THREADS={num_threads}",
                f"OMP_NUM_THREADS={num_threads}",
                f"NUMEXPR_NUM_THREADS={num_threads}",
            ],
        )
        return "\n".join(f"export {arg}" for arg in extra_env_vars)

    def extra_script(self, *, index: int | None = None) -> str:
        """Script that will be run before the main scheduler."""
        assert self._extra_script is not None
        if self.single_job_script:
            assert isinstance(self._extra_script, str)
            return self._extra_script
        assert index is not None
        assert isinstance(self._extra_script, tuple), self._extra_script
        return _maybe_call(self._extra_script[index])

    def write_job_script(
        self,
        name: str,
        options: dict[str, Any],
        index: int | None = None,
    ) -> None:
        """Writes a job script."""
        with self.batch_fname(name).open("w", encoding="utf-8") as f:
            job_script = self.job_script(options, index=index)
            f.write(job_script)

    def _multi_job_script_options(self, index: int) -> dict[str, Any]:
        assert self._command_line_options is not None
        assert isinstance(self.cores, tuple)
        options = dict(self._command_line_options)  # copy
        executor_type = self._get_executor_type(index=index)
        options["--executor-type"] = executor_type
        options["--n"] = self._get_cores(index=index)
        if executor_type == "ipyparallel":
            options["--n"] -= 1
        return options

    def start_job(self, name: str, *, index: int | None = None) -> None:
        """Writes a job script and submits it to the scheduler."""
        if not self.single_job_script:
            assert index is not None
            options = self._multi_job_script_options(index)
            self.write_job_script(name, options, index=index)

        submit_cmd = f"{self.submit_cmd} {name} {self.batch_fname(name)}"
        run_submit(submit_cmd)

    def __getstate__(self) -> dict[str, Any]:
        """Return the state of the scheduler."""
        return {
            "cores": self.cores,
            "python_executable": self.python_executable,
            "log_folder": self.log_folder,
            "mpiexec_executable": self.mpiexec_executable,
            "executor_type": self.executor_type,
            "num_threads": self.num_threads,
            "extra_scheduler": self._extra_scheduler,
            "extra_env_vars": self._extra_env_vars,
            "extra_script": self._extra_script,
        }

    def __setstate__(self, state: dict[str, Any]) -> None:
        """Set the state of the scheduler."""
        self.__init__(**state)  # type: ignore[misc]


def _maybe_call(obj: Any) -> Any:
    if callable(obj):
        return obj()
    return obj



"""Tests for the SLURM scheduler."""

from __future__ import annotations

import textwrap
from pathlib import Path
from unittest.mock import patch

import pytest

import adaptive_scheduler
from adaptive_scheduler._scheduler.slurm import SLURM

from .helpers import PARTITIONS, temporary_working_directory


def test_init_cores() -> None:
    """Test that the cores are set correctly."""
    with pytest.raises(
        ValueError,
        match="Specify either `nodes` and `cores_per_node`, or only `cores`, not both.",
    ):
        SLURM(cores=4, nodes=2, cores_per_node=2)


@pytest.mark.usefixtures("_mock_slurm_partitions")
def test_init_partition() -> None:
    """Test that the partition is set correctly."""
    with pytest.raises(ValueError, match="Invalid partition: nonexistent"):
        SLURM(cores=4, partition="nonexistent")


def test_init_nodes_cores_per_node() -> None:
    """Test that the nodes and cores_per_node are set correctly."""
    with pytest.raises(
        ValueError,
        match="Specify either `nodes` and `cores_per_node`, or only `cores`, not both.",
    ):
        SLURM(nodes=2)


def test_slurm_scheduler() -> None:
    """Test that the slurm scheduler is set correctly."""
    s = SLURM(cores=4)
    assert s._cores == 4
    assert s.nodes is None
    assert s.cores_per_node is None


@pytest.mark.usefixtures("_mock_slurm_partitions")
def test_slurm_scheduler_nodes_cores_per_node() -> None:
    """Test that the nodes and cores_per_node are set correctly."""
    s = SLURM(nodes=2, cores_per_node=2, partition="nc24-low")
    assert s.partition == "nc24-low"
    assert s._cores is None
    assert s.cores == 4
    assert s.nodes == 2
    assert s.cores_per_node == 2


def test_getstate_setstate() -> None:
    """Test that the getstate and setstate methods work."""
    s = SLURM(cores=4)
    state = s.__getstate__()
    s2 = SLURM(cores=2)
    s2.__setstate__(state)
    assert s._cores == s2._cores


def test_job_script() -> None:
    """Test the SLURM.job_script method."""
    s = SLURM(cores=4)
    job_script = s.job_script(options={})
    assert "#SBATCH --ntasks 4" in job_script
    assert "#SBATCH --no-requeue" in job_script
    assert "MKL_NUM_THREADS=1" in job_script
    assert "OPENBLAS_NUM_THREADS=1" in job_script
    assert "OMP_NUM_THREADS=1" in job_script
    assert "NUMEXPR_NUM_THREADS=1" in job_script


@pytest.mark.usefixtures("_mock_slurm_partitions")
def test_start_job(tmp_path: Path) -> None:
    """Test the SLURM.start_job method."""
    s = SLURM(cores=4, partition="nc24-low")
    with (
        temporary_working_directory(tmp_path),
        patch(
            "adaptive_scheduler._scheduler.slurm.run_submit",
        ) as mock_submit,
    ):
        s.write_job_script("testjob", {})
        s.start_job("testjob")
        mock_submit.assert_called_once_with(
            f"sbatch --job-name testjob --output {tmp_path}/testjob-%A.out {tmp_path}/testjob.sbatch",
            "testjob",
        )
        sbatch_file = tmp_path / "testjob.sbatch"
        assert Path.cwd() / s.batch_fname("testjob") == sbatch_file
        assert sbatch_file.exists()
        lines = sbatch_file.read_text()
        assert "#SBATCH --ntasks 4" in lines


@pytest.mark.usefixtures("_mock_slurm_partitions")
def test_slurm_job_script_default() -> None:
    """Test the SLURM.job_script method with default arguments."""
    """Test the SLURM.job_script method with default arguments."""
    s = SLURM(cores=4, partition="nc24-low")
    job_script = s.job_script(options={})

    assert "#SBATCH --ntasks 4" in job_script
    assert "#SBATCH --no-requeue" in job_script
    assert "export MKL_NUM_THREADS=1" in job_script
    assert "export OPENBLAS_NUM_THREADS=1" in job_script
    assert "export OMP_NUM_THREADS=1" in job_script
    assert "export NUMEXPR_NUM_THREADS=1" in job_script


@pytest.mark.usefixtures("_mock_slurm_partitions")
def test_slurm_job_script_custom() -> None:
    """Test the SLURM.job_script method with custom arguments."""
    extra_scheduler = ["--time=1", "--mem=4G"]
    extra_env_vars = ["TMPDIR='/scratch'", "PYTHONPATH='my_dir:$PYTHONPATH'"]
    extra_script = "echo 'test'"

    s = SLURM(
        cores=4,
        partition="nc24-low",
        extra_scheduler=extra_scheduler,
        extra_env_vars=extra_env_vars,
        extra_script=extra_script,
    )
    job_script = s.job_script(options={})

    # Check extra_scheduler
    for opt in extra_scheduler:
        assert f"#SBATCH {opt}" in job_script

    # Check extra_env_vars
    for var in extra_env_vars:
        assert f"export {var}" in job_script

    # Check extra_script
    assert extra_script in job_script

    # Check partition
    assert "#SBATCH --partition=nc24-low" in job_script


@pytest.mark.usefixtures("_mock_slurm_partitions_output")
def test_slurm_partitions() -> None:
    """Test slurm_partitions function."""
    partitions = adaptive_scheduler._scheduler.slurm.slurm_partitions(with_ncores=False)
    assert partitions == [
        "nc24-low",
        "hb120v2-low",
        "hb60-high",
        "nd40v2-mpi",
    ]
    partitions = adaptive_scheduler._scheduler.slurm.slurm_partitions(with_ncores=True)
    assert partitions == PARTITIONS


@pytest.mark.usefixtures("_mock_slurm_partitions")
def test_slurm_partitions_mock() -> None:
    """Test slurm_partitions function."""
    assert adaptive_scheduler._scheduler.slurm.slurm_partitions() == PARTITIONS


def test_slurm_scheduler_job_script_ipyparallel() -> None:
    """Test the SLURM.job_script method."""
    s = SLURM(
        cores=4,
        extra_scheduler=["--exclusive=user", "--time=1"],
        extra_env_vars=["TMPDIR='/scratch'", "PYTHONPATH='my_dir:$PYTHONPATH'"],
        extra_script="echo 'YOLO'",
        executor_type="ipyparallel",
    )
    job_script = s.job_script(options={"--n": 3})
    log_fname = s.log_fname("${NAME}")
    assert (
        job_script.strip()
        == textwrap.dedent(
            f"""\
        #!/bin/bash
        #SBATCH --ntasks 4
        #SBATCH --no-requeue
        #SBATCH --exclusive=user
        #SBATCH --time=1

        export TMPDIR='/scratch'
        export PYTHONPATH='my_dir:$PYTHONPATH'
        export EXECUTOR_TYPE=ipyparallel
        export MKL_NUM_THREADS=1
        export OPENBLAS_NUM_THREADS=1
        export OMP_NUM_THREADS=1
        export NUMEXPR_NUM_THREADS=1

        echo 'YOLO'

        profile=adaptive_scheduler_${{SLURM_JOB_ID}}

        echo "Creating profile ${{profile}}"
        ipython profile create ${{profile}}

        echo "Launching controller"
        ipcontroller --ip="*" --profile=${{profile}} --log-to-file &
        sleep 10

        echo "Launching engines"
        srun --ntasks 3 ipengine \\
            --profile=${{profile}} \\
            --cluster-id='' \\
            --log-to-file &

        echo "Starting the Python script"
        srun --ntasks 1 {s.python_executable} {s.launcher} \\
            --profile ${{profile}} \\
            --log-fname {log_fname} \\
            --job-id ${{SLURM_JOB_ID}} \\
            --name ${{NAME}} \\
            --n 3
        """,
        ).strip()
    )


@pytest.mark.usefixtures("_mock_slurm_partitions")
def test_slurm_scheduler_ipyparallel() -> None:
    """Test the SLURM.job_script method."""
    s = SLURM(
        extra_scheduler=["--exclusive=user", "--time=1"],
        extra_env_vars=["TMPDIR='/scratch'", "PYTHONPATH='my_dir:$PYTHONPATH'"],
        extra_script="echo 'YOLO'",
        executor_type="ipyparallel",
        nodes=999,
        cores_per_node=24,
        partition="nc24-low",
        exclusive=True,
    )
    assert s.cores == 999 * 24
    ipy = s._executor_specific("TEST", {"--n": 23975})
    log_fname = s.log_fname("TEST")
    print(ipy)
    assert (
        ipy.strip()
        == textwrap.dedent(
            f"""\
        profile=adaptive_scheduler_${{SLURM_JOB_ID}}

        echo "Creating profile ${{profile}}"
        ipython profile create ${{profile}}

        echo "Launching controller"
        ipcontroller --ip="*" --profile=${{profile}} --log-to-file &
        sleep 10

        echo "Launching engines"
        srun --ntasks 23975 ipengine \\
            --profile=${{profile}} \\
            --cluster-id='' \\
            --log-to-file &

        echo "Starting the Python script"
        srun --ntasks 1 {s.python_executable} {s.launcher} \\
            --profile ${{profile}} \\
            --log-fname {log_fname} \\
            --job-id ${{SLURM_JOB_ID}} \\
            --name TEST \\
            --n 23975
        """,
        ).strip()
    )


def test_multiple_jobs() -> None:
    """Test that multiple jobs can be started."""
    cores = (3, 4, 5)
    s = SLURM(cores=cores)
    for i, n in enumerate(cores):
        js = s.job_script(options={}, index=i)
        assert f"#SBATCH --ntasks {n}" in js
        assert js.count("--ntasks") == 1
    assert isinstance(s._extra_scheduler, tuple)
    assert len(s._extra_scheduler) == 3

    s = SLURM(cores_per_node=cores, nodes=2)
    assert isinstance(s.nodes, tuple)
    assert s.cores == tuple(2 * n for n in cores)

    partitions = ("nc24-low", "hb120v2-low")
    with patch("adaptive_scheduler._scheduler.slurm.slurm_partitions") as mock:
        mock.return_value = {"nc24-low": 24, "hb120v2-low": 120}

        s = SLURM(partition=partitions, cores=1)
        assert s.cores == (1, 1)
        for i, p in enumerate(partitions):
            js = s.job_script(options={}, index=i)
            assert f"#SBATCH --partition={p}" in js
            assert js.count("--partition") == 1

    s = SLURM(cores=cores, extra_scheduler=["--time=1"], executor_type="ipyparallel")
    js = s.job_script(options={}, index=0)
    assert "#SBATCH --time=1" in js
    assert "--ntasks 2" in js

    # Check with incorrect types
    with pytest.raises(TypeError, match=r"Expected `<class 'int'>`"):
        SLURM(cores="4")  # type: ignore[arg-type]

    with pytest.raises(ValueError, match="All tuples should have the same length."):
        SLURM(cores_per_node=(4,), nodes=(2, 2))  # type: ignore[arg-type]

    s = SLURM(cores=1, exclusive=(True, False))
    js = s.job_script(options={}, index=0)
    opt = "#SBATCH --exclusive"
    assert opt in js
    js = s.job_script(options={}, index=1)
    assert opt not in js

    s = SLURM(cores=1, exclusive=True)
    js = s.job_script(options={})
    assert opt in js

    s = SLURM(cores=1, extra_env_vars=(["YOLO=1"], []))
    js = s.job_script(options={}, index=0)
    assert "export YOLO=1" in js
    js = s.job_script(options={}, index=1)
    assert "export YOLO=1" not in js

    s = SLURM(cores=1, extra_script=("echo 'YOLO'", ""))
    js = s.job_script(options={}, index=0)
    print(js)
    assert "echo 'YOLO'" in js
    js = s.job_script(options={}, index=1)
    assert "echo 'YOLO'" not in js

    s = SLURM(cores=1, num_threads=(1, 2))
    js = s.job_script(options={}, index=0)
    print(js)
    assert "export OPENBLAS_NUM_THREADS=1" in js
    js = s.job_script(options={}, index=1)
    assert "export OPENBLAS_NUM_THREADS=2" in js


def test_multi_job_script_options() -> None:
    """Test the SLURM.job_script method."""
    s = SLURM(cores=2, executor_type=("ipyparallel", "sequential"))
    assert not s.single_job_script
    s._command_line_options = {"--n": 2}

    # Test ipyparallel
    with patch("adaptive_scheduler._scheduler.slurm.run_submit") as mock_submit:
        s.start_job("testjob", index=0)
        mock_submit.assert_called_once()
        args, _ = mock_submit.call_args
        assert args[0].startswith("sbatch")
    assert "--executor-type ipyparallel" in s.batch_fname("testjob").read_text()

    # Test sequential
    with patch("adaptive_scheduler._scheduler.slurm.run_submit") as mock_submit:
        s.start_job("testjob", index=1)
        mock_submit.assert_called_once()
        args, _ = mock_submit.call_args
        assert args[0].startswith("sbatch")
    assert "--executor-type sequential" in s.batch_fname("testjob").read_text()


def test_not_changing_attributes() -> None:
    """Tests bug fixed in https://github.com/basnijholt/adaptive-scheduler/pull/215."""
    s = SLURM(cores_per_node=100, nodes=(1, 2))
    assert s._extra_env_vars == ([], [])
    js0 = s.job_script(options={}, index=0)
    assert s._extra_env_vars == ([], [])
    js1 = s.job_script(options={}, index=1)
    assert s._extra_env_vars == ([], [])
    assert js0.count("MKL_NUM_THREADS") == 1
    assert js1.count("MKL_NUM_THREADS") == 1


def test_callable_scheduler_arguments() -> None:
    """Test that the scheduler arguments can be callables."""
    s = SLURM(
        cores=(4, lambda: 2),
        executor_type=(lambda: "ipyparallel", "mpi4py"),
        num_threads=(lambda: 2, 1),
        extra_scheduler=(
            ["--exclusive=user", "--time=1"],
            lambda: ["--exclusive=user", "--time=2"],
        ),
        extra_env_vars=(lambda: ["from=func"], ["from=static"]),
        extra_script=(lambda: "echo 'func'", "echo 'static'"),
        exclusive=(lambda: True, lambda: False),
    )

    js0 = s.job_script(options={}, index=0)
    js1 = s.job_script(options={}, index=1)
    assert js0 != js1
    extra_scheduler0 = s.extra_scheduler(index=0)
    extra_scheduler1 = s.extra_scheduler(index=1)
    assert extra_scheduler0 == "#SBATCH --exclusive\n#SBATCH --exclusive=user\n#SBATCH --time=1"
    assert extra_scheduler1 == "#SBATCH --exclusive=user\n#SBATCH --time=2"
    extra_env_vars0 = s.extra_env_vars(index=0)
    extra_env_vars1 = s.extra_env_vars(index=1)
    assert extra_env_vars0.startswith("export from=func\n")
    assert extra_env_vars1.startswith("export from=static\n")
    extra_script0 = s.extra_script(index=0)
    extra_script1 = s.extra_script(index=1)
    assert extra_script0 == "echo 'func'"
    assert extra_script1 == "echo 'static'"
    cores0 = s._get_cores(index=0)
    cores1 = s._get_cores(index=1)
    assert cores0 == 4
    assert cores1 == 2
    executor_type0 = s._get_executor_type(index=0)
    executor_type1 = s._get_executor_type(index=1)
    assert executor_type0 == "ipyparallel"
    assert executor_type1 == "mpi4py"


def test_callable_scheduler_arguments_with_cores_per_node(_mock_slurm_partitions) -> None:  # noqa: PT019, ANN001
    """Test that the scheduler arguments can be callables."""
    s = SLURM(
        cores_per_node=(4, lambda: 2, 3),
        nodes=(lambda: 2, lambda: 1, 3),
        partition=(lambda: "nc24-low", lambda: "hb120v2-low", "hb60-high"),
        executor_type=(lambda: "ipyparallel", "mpi4py", "mpi4py"),
        exclusive=True,
    )
    assert isinstance(s.cores, tuple)
    assert callable(s.cores[0])
    assert s.cores[0]() == 8
    assert callable(s.cores[1])
    assert s.cores[1]() == 2
    assert not callable(s.cores[2])
    assert s.cores[2] == 9

    js0 = s.job_script(options={}, index=0)
    js1 = s.job_script(options={}, index=1)
    js2 = s.job_script(options={}, index=2)
    assert js0 != js1
    assert js0 != js2
    assert "--partition=nc24-low" in js0
    assert "--partition=hb120v2-low" in js1
    assert "--partition=hb60-high" in js2
    assert "--ntasks 8" in js0
    assert "--ntasks 2" in js1
    assert "--ntasks 9" in js2

    ex0, start0 = s._ipyparallel(index=0)
    ex1, start1 = s._ipyparallel(index=1)
    ex2, start2 = s._ipyparallel(index=2)
    assert ex0 != ex1
    assert ex0 != ex2
    assert start0 == start1
    assert start0 == start2
    assert start0 == ("    --profile ${profile}",)
